#include "Entity3dCollection.h"

Entity3dCollection::Entity3dCollection(void)
{

	Clear();
}

Entity3dCollection::~Entity3dCollection(void)
{

}

void Entity3dCollection::AddEntity(Entity3d ent, int x, int y, int textureIndex)
{
	assert(x >= 0);
	assert(y >= 0);
	assert(x < MAX_ENTITIES);
	assert(y < MAX_ENTITIES);

	if(textureIndex >= 0 && textureIndex < TEXTURE_COUNT)
	{
		m_entities[textureIndex][x][y] = ent;

		//connect naturally-connected nodes
		ConnectionStruct connections;
		DiscoverConnectedNodes(textureIndex, x, y, connections);

		if((connections.east) && (ent.Connectivity.Connections.east))
			UnionTwoNodes(x, y, textureIndex, 1, 0);
		if((connections.west) && (ent.Connectivity.Connections.west))
			UnionTwoNodes(x, y, textureIndex, -1, 0);
		if((connections.north) && (ent.Connectivity.Connections.north))
			UnionTwoNodes(x, y, textureIndex, 0, 1);
		if((connections.south) && (ent.Connectivity.Connections.south))
			UnionTwoNodes(x, y, textureIndex, 0, -1);
	}
}

void Entity3dCollection::UnionTwoNodes(int x, int y, int textureIndex, int deltax, int deltay)
{
	m_disjointSets[textureIndex].Union(
		m_disjointSets[textureIndex].FindSet(x + MAX_ENTITIES * y),
		m_disjointSets[textureIndex].FindSet(x + deltax + MAX_ENTITIES * (y + deltay)));
}

bool Entity3dCollection::AreDisjoint(int x, int y, int textureIndex, int deltax, int deltay)
{
	return 
		m_disjointSets[textureIndex].FindSet(x + MAX_ENTITIES * y) !=
		m_disjointSets[textureIndex].FindSet(x + deltax + MAX_ENTITIES * (y + deltay));
}

//Run this after building the connection.
//Naturally-connected nodes will already exist in the disjoint sets structure (unioned together).
void Entity3dCollection::EnsureAdjacentConnectivity(int textureIndex)
{
	for(int x = 0; x < MAX_ENTITIES; x++)
	{
		for(int y = 0; y < MAX_ENTITIES; y++)
		{
			if(m_entities[textureIndex][x][y].rendered_type == RENDERED_NONE)
				continue;

			if(
				(x > 0)
				&& (m_entities[textureIndex][x-1][y].rendered_type != RENDERED_NONE)  //other node exists
				&& (AreDisjoint(x, y, textureIndex, -1, 0))							  //other node disjoint
			  )
			{
				//connect.
				m_entities[textureIndex][x-1][y].Connectivity.Connections.east = 1;
				m_entities[textureIndex][x][y].Connectivity.Connections.west = 1;
				UnionTwoNodes(x, y, textureIndex, -1, 0);
			}

			if(
				(x < MAX_ENTITIES -1)
				&& (m_entities[textureIndex][x+1][y].rendered_type != RENDERED_NONE)  //other node exists
				&& (AreDisjoint(x, y, textureIndex, 1, 0))							  //other node disjoint
			  )
			{
				//connect.
				m_entities[textureIndex][x+1][y].Connectivity.Connections.west = 1;
				m_entities[textureIndex][x][y].Connectivity.Connections.east = 1;
				UnionTwoNodes(x, y, textureIndex, 1, 0);
			}

			if(
				(y > 0)
				&& (m_entities[textureIndex][x][y-1].rendered_type != RENDERED_NONE)  //other node exists
				&& (AreDisjoint(x, y, textureIndex, 0, -1))							  //other node disjoint
			  )
			{
				//connect.
				m_entities[textureIndex][x][y-1].Connectivity.Connections.north = 1;
				m_entities[textureIndex][x][y].Connectivity.Connections.south = 1;
				UnionTwoNodes(x, y, textureIndex, 0, -1);
			}

			if(
				(y < MAX_ENTITIES - 1)
				&& (m_entities[textureIndex][x][y+1].rendered_type != RENDERED_NONE)  //other node exists
				&& (AreDisjoint(x, y, textureIndex, 0, 1))							  //other node disjoint
			  )
			{
				//connect.
				m_entities[textureIndex][x][y+1].Connectivity.Connections.south = 1;
				m_entities[textureIndex][x][y].Connectivity.Connections.north = 1;
				UnionTwoNodes(x, y, textureIndex, 0, +1);
			}
		}
	}
}

void Entity3dCollection::Clear()
{
	for(int i = 0; i < TEXTURE_COUNT; i++)
	{
		m_disjointSets[i].Reset();

		for(int j = 0; j < MAX_ENTITIES; j++)
		{
			for(int k = 0; k < MAX_ENTITIES; k++)
			{
				m_entities[i][j][k].rendered_type = RENDERED_NONE;
			}
		}
	}
}

void Entity3dCollection::CompileDisplayLists(GLuint firstEntityList)
{
	for(int i = 0; i < TEXTURE_COUNT; i++)
	{
		//build display list
		glNewList(firstEntityList + i, GL_COMPILE);

		for(int j = 0; j < MAX_ENTITIES; j++)
		{
			for(int k = 0; k < MAX_ENTITIES; k++)
			{
				RenderSingleNode(i, j, k);
			}
		}

		glEndList();
	}
}

void Entity3dCollection::Render(GLuint *textures, GLuint firstEntityList)
{
	for(int i = 0; i < TEXTURE_COUNT; i++)
	{
		glBindTexture(GL_TEXTURE_2D, textures[i]);
		glCallList(firstEntityList + i);
	}
}


void Entity3dCollection::RenderSingleNode(int texture, int x, int y)
{
	if(m_entities[texture][x][y].rendered_type == RENDERED_NONE)
		return;

	ConnectionStruct connections;
	DiscoverConnectedNodes(texture, x, y, connections);

	m_entities[texture][x][y].Render(connections);
}

void Entity3dCollection::DiscoverConnectedNodes(int texture, int x, int y, ConnectionStruct &connections)
{
	connections.east = connections.north = connections.south = connections.west = 0;

	//some neighbors are not empty.
	if(x > 0)
		connections.west = (m_entities[texture][x - 1][y].rendered_type != RENDERED_NONE) && (m_entities[texture][x - 1][y].Connectivity.Connections.east);
	if(x < MAX_ENTITIES - 1)
		connections.east = (m_entities[texture][x + 1][y].rendered_type != RENDERED_NONE) && (m_entities[texture][x + 1][y].Connectivity.Connections.west);
	if(y > 0)
		connections.south = (m_entities[texture][x][y - 1].rendered_type != RENDERED_NONE) && (m_entities[texture][x][y - 1].Connectivity.Connections.north);
	if(y < MAX_ENTITIES - 1)
		connections.north = (m_entities[texture][x][y + 1].rendered_type != RENDERED_NONE) && (m_entities[texture][x][y + 1].Connectivity.Connections.south);
}