#include "ProgramManager.h"
#include <utils.h>
#include <math.h>

#include <VersionInfo.h>


#include "NoiseSlice.h"

#define PROGRAM_NAME "NoiseViewer v" VERSION_STRING

#define FPS 60
#define MS_PER_FRAME 1000 / FPS

static void _Sleep(int ms)
{
    // [WODE] *cough*
#ifdef __APPLE__
    sleep(ms/1000);
#else
    Sleep(ms);
#endif
    
}

ProgramManager* ProgramManager::m_Manager = NULL;

ProgramManager::ProgramManager(void)
{
	m_Manager = this;
	m_MouseState = MOUSE_INACTIVE;
	mousedown_x = 0;
	mousedown_y = 0;
	width = height = 0;

	m_window_id = 0;
}

ProgramManager::~ProgramManager(void)
{
	m_Manager = NULL;
}


void ProgramManager::mousebutton(int button, int state, int x, int y)
{
	GLUI_String message("");

	if(state != GLUT_DOWN)
		m_Manager->m_MouseState = MOUSE_INACTIVE;
	else
	{
		m_Manager->mousedown_x = x;
		m_Manager->mousedown_y = y;

		if(button == GLUT_RIGHT_BUTTON)
		{
			m_Manager->m_MouseState = MOUSE_ZOOMING;
			message = "Move mouse to change zoom level of the current layer";
		}
		else if(button == GLUT_LEFT_BUTTON)
		{

			if(glutGetModifiers() & GLUT_ACTIVE_SHIFT)
			{
				m_Manager->m_MouseState = MOUSE_CUTOFF;
				message = "Move mouse to adjust parameter of the current layer";
			}
			else if(glutGetModifiers() & GLUT_ACTIVE_CTRL)
			{
				m_Manager->m_MouseState = MOUSE_ZOOMING;
				message = "Move mouse to change zoom level of the current layer";
			}
			else
			{
				m_Manager->m_MouseState = MOUSE_PANNING;
				message = "Move mouse to pan the current layer";
			}
		}
	}

	m_Manager->m_logEdit->set_text(message);
}

void ProgramManager::KeyboardDown(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 27:
		LOG(L"righty ho");
		break;
	}

	m_Manager->theWorld->KeyDown[key] = 1;
}

void ProgramManager::KeyboardUp(unsigned char key, int x, int y)
{
	m_Manager->theWorld->KeyDown[key] = 0;
}

void ProgramManager::doMouseMove(int x, int y)
{
	if(m_MouseState != MOUSE_INACTIVE)
	{
		int delta_x = mousedown_x - x;
		int delta_y = (mousedown_y - y);
		mousedown_x = x;
		mousedown_y = y;

		switch(m_MouseState)
		{
		case MOUSE_PANNING:
			PanMouse(delta_x, delta_y);
			break;
		case MOUSE_ZOOMING:
			ZoomMouse(delta_x, delta_y);
			break;
		case MOUSE_CUTOFF:
			m_layerManager.IncrementParameter(((float)delta_y)/100);
			break;
		}
	}
}

void ProgramManager::PanMouse(int delta_x, int delta_y)
{
	//get relative to size of window
	float dx, dy;
	dx = (float)delta_x / width;
	dy = (float)delta_y / height;

	switch(m_layerManager.m_renderMode)
	{
	case DISPLAY_2D:
		m_layerManager.Translate(dx, dy);
		break;
	case DISPLAY_3D:
		theWorld->PanMouse(dx, dy);
		m_layerManager.UpdateTextures();
		break;
	}

	glutSetWindow(m_window_id);
	glutPostRedisplay();
}

void ProgramManager::ZoomMouse(int delta_x, int delta_y)
{

	//get relative to size of window
	float dx, dy;
	dx = (float)delta_x / width;
	dy = (float)delta_y / height;

	switch(m_layerManager.m_renderMode)
	{
	case DISPLAY_2D:
		m_layerManager.Scale(dy, dy);
		break;
	case DISPLAY_3D:
		theWorld->ZoomMouse(dy);
		m_layerManager.UpdateTextures();
		break;
	}

	glutSetWindow(m_window_id);
	glutPostRedisplay();
}

void ProgramManager::mousemove(int x, int y)
{
	m_Manager->doMouseMove(x, y);
}

void ProgramManager::tick()
{
	int ms = (int)m_Manager->m_timer.TimeElapsed();
	//if last frame was fast, rest a bit
	if(ms < MS_PER_FRAME)
    {
		_Sleep(MS_PER_FRAME - ms);
    }

	//perform changes
	bool worldChanged = false;
	if(m_Manager->m_layerManager.m_renderMode == DISPLAY_3D)
		worldChanged = m_Manager->theWorld->DoFrame(ms);

	if(m_Manager->m_layerManager.ShouldUpdateLayers() || worldChanged)
	{
		m_Manager->m_layerManager.UpdateTextures();

		//must set main GLUT window
		glutSetWindow(m_Manager->m_window_id);
		glutPostRedisplay();
	}
}


void ProgramManager::display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT);	// Clear The Screen And The Depth Buffer
	//glLoadIdentity();									// Reset The View

	m_Manager->SetCamera();

	m_Manager->m_layerManager.Render();

	glFlush();
	glutSwapBuffers();
}

void ProgramManager::reshape(int w, int h)
{
	int x,y;
	GLUI_Master.get_viewport_area(&x, &y, &m_Manager->width, &m_Manager->height);

	GLUI_Master.auto_set_viewport();
}

void ProgramManager::SetCamera()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	switch(m_layerManager.m_renderMode)
	{
		case DISPLAY_2D:
				glOrtho(0, 1, 0, 1, -1, 1);
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();

				EnableLighting(false);
			break;

		case DISPLAY_3D:
				EnableLighting(true);
				gluPerspective(60, (GLfloat)width / (GLfloat)height, 0.01, 100.0);
				theWorld->CreateView();
				//EnableLighting(true);
			break;
	}
}

void ProgramManager::EnableLighting(bool enable)
{
	if(enable)
	{
		theWorld->BuildLighting();
	}
	else
	{
		glDisable(GL_LIGHTING);
		glDisable(GL_LIGHT0);
		glDisable(GL_COLOR_MATERIAL);
		
	}
}

int ProgramManager::Run(int argc, char **argv)
{
	int x, y;
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
	glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);

	theWorld = new World();

	m_window_id = glutCreateWindow(PROGRAM_NAME);

	glEnable(GL_TEXTURE_2D);							// Enable Texture Mapping ( NEW )
	//glEnable(GL_BLEND);
	//glBlendFunc (GL_SRC_ALPHA, GL_DST_ALPHA);
	glShadeModel(GL_SMOOTH);							// Enable Smooth Shading
	glClearDepth(1.0f);									// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
	glDepthFunc(GL_LESS);								// The Type Of Depth Testing To Do
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations


	//set up key state tracking
	glutIgnoreKeyRepeat(true);
	GLUI_Master.set_glutKeyboardFunc(ProgramManager::KeyboardDown);
	glutKeyboardUpFunc(ProgramManager::KeyboardUp);

	//if you need to use this function, don't go through GLUT, use this:
	//GLUI_Master.set_glutSpecialFunc( myGlutSpecial );

	GLUI_Master.set_glutMouseFunc( ProgramManager::mousebutton );
	GLUI_Master.set_glutReshapeFunc( ProgramManager::reshape );

	//glutTimerFunc(10, ProgramManager::tick, 0);
	glutMotionFunc(ProgramManager::mousemove);
	glutDisplayFunc(ProgramManager::display);

	GLUI *glui = GLUI_Master.create_glui_subwindow(m_window_id, GLUI_SUBWINDOW_RIGHT);
	glui->set_main_gfx_window(m_window_id);

	GLUI_Master.set_glutIdleFunc(ProgramManager::tick);
	GLUI *log_window = GLUI_Master.create_glui_subwindow(m_window_id, GLUI_SUBWINDOW_BOTTOM);
	log_window->set_main_gfx_window(m_window_id);
	m_logEdit = log_window->add_edittext("Info: ");
	m_logEdit->set_alignment(GLUI_ALIGN_LEFT);
	m_logEdit->set_w(400);

	m_layerManager.Initialize(glui);

	GLUI_Master.get_viewport_area(&x, &y, &width, &height);
	m_layerManager.SetViewportSize(width, height);

	m_layerManager.SetDefaults();

	glutMainLoop();

	delete theWorld;
	return 0;
}